function [M,M0,de76,de00,k,output] = estimateMatrix(raw,xyzRef,xyzn)
    % Determines Compromise Color Matrix M from CC24 raw capture.
    % Solves overdetermined Ax=b for x -> M: raw * M = xyzRef, Nx3 data
    % If raw data is white balanced, M' TRANSPOSE corresponds to DNG 
    % Forward Matrix FM for given illuminant.  By Jack Hogan, ver 0.25
    % See www.strollswithmydog.com/determining-forward-color-matrix/
    % and www.dpreview.com/forums/post/65552794
    % INPUTS:
    % raw 24x3, CC24 RGB raw data captured under illuminant
    % xyzRef 24x3, known reference XYZ values for CC24 under illuminant
    % xyzn 1x3, XYZ coords of illuminant at time of capture
    % OUTPUTS:
    % M 3x3, the resulting Compromise Color Matrix so that XYZ = raw * M
    % k, Y luminance gain introduced by the matrix, xyzRef = raw * M * k
    % output, output message of search algorithm.
    
    xyzn = xyzn / xyzn(2);                          %illuminant White Point
    xyzRef = xyzRef / xyzRef(22,2) * 0.19;          %norm reference data
    raw = raw ./ raw(22,2) * xyzRef(22,2);          %norm, intensity only
    labRef = xyz2labn(xyzRef,xyzn);                 %CC24 lab reference
    
    M0 = inv(raw' * raw) * raw' * xyzRef;           %#ok<MINV> seed matrix
    
    opts = optimset('MaxFunEvals',10^5,'MaxIter',10^5,'Display','off');
    sse = @(M) sum(deltaE00( xyz2labn(raw*M,xyzn),labRef )); %or other metric
    [M,~,~,output] = fminunc(sse,M0,opts);            %or fminunc
    
    k = sum(M(:,2));                                %Y gain
    M = M / k;                                      %normalize sum(Y)=1
                                                    %remember M is transpose of FM
    %% Output - needs a couple of extra functions at bottom, ignore
    de76 = deltaE76( xyz2labn(raw*M*k,xyzn),labRef ); %color differences to
    de00 = deltaE00( xyz2labn(raw*M*k,xyzn),labRef ); %reference data
    
    disp('raw * M = xyz, with raw Nx3.  Transpose of M = DNG Forward Matrix:')
    disp(num2str(M')); disp('-------------------------------');
    disp(num2str(sum(M)));
    cctDuv(sum(M));                                 %optional
    disp(['Mean over 24 patches: de00 = ' num2str(mean(de00))])
    disp(['de76 = ' ...
        num2str(mean(de76)) ', SMI = ' num2str(100-5.5*mean(de76(1:18)),3)...
        ', k = ' num2str(k) '.'])
    
    figure; bar3c(reshape(de00,6,[])',roo2disp(colorchecker));  %requires optprop
    zlabel('\DeltaE_{00}'); zlim([0 4])
end
function lab = xyz2labn(xyz,xyzn)
    % Converts Nx3 xyz values to Nx3 L*ab given reference white coords xyzn
    
    fx = f( xyz(:,1)/xyzn(1) );
    fy = f( xyz(:,2)/xyzn(2) );
    fz = f( xyz(:,3)/xyzn(3) );
    
    lab = [116*fy-16, 500*(fx-fy), 200*(fy - fz)];
    
    function y = f(t)
        y = zeros(size(t));
        gt = t > (6/29)^3;
        y(gt) = t(gt).^(1/3);
        y(~gt) =  t(~gt)/3/(6/29)^2 + 4/29;
    end
end
%%
function de76 = deltaE76(lab1,lab2)                           
    %Calculates deltaE76 between the two arrays in Nx3 lab notation
    de76 = sqrt(sum((lab2-lab1).^2,2));
end
%%
function de00 = deltaE00(Lab1, Lab2) 
    % deltaE00: Calculates the CIE Delta E00 Color Difference
    % Original by Lawrence Taplin 08Jan01, adapted to Nx3 and Delta Values
    % made absolute per Optprop
    % Note: SL,SC&SH (bottom) are based on the second set of Lab values.
    
    %CIELAB Chroma
    C1 = sqrt(Lab1(:,2).^2+Lab1(:,3).^2);
    C2 = sqrt(Lab2(:,2).^2+Lab2(:,3).^2);
    
    %Lab Prime
    mC = (C1+C2)./2;
    G=0.5*(1-sqrt((mC.^7)./((mC.^7)+(25.^7))));
    LabP1 = [Lab1(:,1)  Lab1(:,2).*(1+G)  Lab1(:,3)];
    LabP2 = [Lab2(:,1)  Lab2(:,2).*(1+G)  Lab2(:,3)];
    
    %Chroma
    CP1 = sqrt(LabP1(:,2).^2+LabP1(:,3).^2);
    CP2 = sqrt(LabP2(:,2).^2+LabP2(:,3).^2);
    
    %Hue Angle
    hP1t = atan2d(LabP1(:,3),LabP1(:,2));
    hP1t = hP1t+(hP1t<0).*360;
    hP2t = atan2d(LabP2(:,3),LabP2(:,2));
    hP2t = hP2t+(hP2t<0).*360;
    
    %Add in 360 to the smaller hue angle if absolute value of difference is > 180
    hP1 = hP1t + ((hP1t<hP2t)&(abs(hP1t-hP2t)>180)).*360;
    hP2 = hP2t + ((hP1t>hP2t)&(abs(hP1t-hP2t)>180)).*360;
    
    %Delta Values
    DLP = abs(LabP1(:,1) - LabP2(:,1));
    DCP = abs(CP1 - CP2);
    DhP = abs(hP1 - hP2);
    DHP = 2*(CP1.*CP2).^(1/2).*sind(DhP./2);
    
    %Arithmetic mean of LCh' values
    mLP = (LabP1(:,1)+LabP2(:,1))./2;
    mCP = (CP1+CP2)./2;
    mhP = (hP1+hP2)./2;
    
    %Weighting Functions
    SL = 1+(0.015.*(mLP-50).^2)./sqrt(20+(mLP-50).^2);
    SC = 1+0.045.*mCP;
    T = 1-0.17.*cosd(mhP-30)+0.24.*cosd(2.*mhP)+0.32.*cosd(3.*mhP+6)-0.2.*cosd(4.*mhP-63);
    SH = 1+0.015.*mCP.*T;
    
    %Rotation function
    RC = 2.*sqrt((mCP.^7)./((mCP.^7)+25.^7));
    DTheta = 30.*exp(-((mhP-275)./25).^2);
    RT = -sind(2.*DTheta).*RC;
    
    %Parametric factors
    kL = 1;
    kC = 1;
    kH = 1;
    
    de00 = ((DLP./kL./SL).^2+(DCP./kC./SC).^2+(DHP./kH./SH).^2+(RT.*(DCP./kC./SC).*(DHP./kH./SH))).^(1/2);
end